@@ -41,6 +41,16 @@
 #include <syslog.h>
 #include <time.h>
 
+#ifdef APP_MACHINEID_SUPPORT
+#include <systemd/sd-id128.h>
+#endif
+
+#ifdef HAVE_VALGRIND
+#include <valgrind.h>
+#else
+#define RUNNING_ON_VALGRIND 0
+#endif
+
 #ifdef __linux__
 #  include <malloc.h>
 #endif
@@ -72,7 +82,7 @@ const char *system_id_from_string(struct cmd_context *cmd, const char *str)
 	}
 
 	if (!strncmp(system_id, "localhost", 9)) {
-		log_warn("WARNING: system ID may not begin with the string \"localhost\".");
+		log_warn("WARNING: System ID may not begin with the string \"localhost\".");
 		return NULL;
 	}
 
@@ -129,9 +139,12 @@ static const char *_read_system_id_from_file(struct cmd_context *cmd, const char
 	return system_id;
 }
 
+/* systemd-id128 new produced: f64406832c2140e8ac5422d1089aae03 */
+#define LVM_APPLICATION_ID SD_ID128_MAKE(f6,44,06,83,2c,21,40,e8,ac,54,22,d1,08,9a,ae,03)
+
 static const char *_system_id_from_source(struct cmd_context *cmd, const char *source)
 {
-	char filebuf[PATH_MAX];
+	char buf[PATH_MAX];
 	const char *file;
 	const char *etc_str;
 	const char *str;
@@ -150,10 +163,25 @@ static const char *_system_id_from_source(struct cmd_context *cmd, const char *s
 		goto out;
 	}
 
+#ifdef APP_MACHINEID_SUPPORT
+	if (!strcasecmp(source, "appmachineid")) {
+		sd_id128_t id = { 0 };
+
+		if (sd_id128_get_machine_app_specific(LVM_APPLICATION_ID, &id) != 0)
+			log_warn("WARNING: sd_id128_get_machine_app_specific() failed %s (%d).",
+				 strerror(errno), errno);
+
+		if (dm_snprintf(buf, PATH_MAX, SD_ID128_FORMAT_STR, SD_ID128_FORMAT_VAL(id)) < 0)
+			stack;
+		system_id = system_id_from_string(cmd, buf);
+		goto out;
+	}
+#endif
+
 	if (!strcasecmp(source, "machineid") || !strcasecmp(source, "machine-id")) {
 		etc_str = find_config_tree_str(cmd, global_etc_CFG, NULL);
-		if (dm_snprintf(filebuf, sizeof(filebuf), "%s/machine-id", etc_str) != -1)
-			system_id = _read_system_id_from_file(cmd, filebuf);
+		if (dm_snprintf(buf, sizeof(buf), "%s/machine-id", etc_str) != -1)
+			system_id = _read_system_id_from_file(cmd, buf);
 		goto out;
 	}
 
@@ -230,7 +258,7 @@ static void _get_sysfs_dir(struct cmd_context *cmd, char *buf, size_t buf_size)
 		return;
 	}
 
-	(void) dm_strncpy(buf, sys_mnt, buf_size);
+	dm_strncpy(buf, sys_mnt, buf_size);
 }
 
 static uint32_t _parse_debug_fields(struct cmd_context *cmd, int cfg, const char *cfgname)
@@ -320,6 +348,33 @@ static int _parse_debug_classes(struct cmd_context *cmd)
 	return debug_classes;
 }
 
+static uint32_t _parse_log_journal(struct cmd_context *cmd, int cfg, const char *cfgname)
+{
+	const struct dm_config_node *cn;
+	const struct dm_config_value *cv;
+	uint32_t fields = 0;
+	uint32_t val;
+
+	if (!(cn = find_config_tree_array(cmd, cfg, NULL))) {
+		log_debug("Unable to find configuration for log/%s.", cfgname);
+		return 0;
+	}
+
+	for (cv = cn->v; cv; cv = cv->next) {
+		if (cv->type != DM_CFG_STRING) {
+			log_verbose("log/%s contains a value which is not a string.  Ignoring.", cfgname);
+			continue;
+		}
+
+		if ((val = log_journal_str_to_val(cv->v.str)))
+			fields |= val;
+		else
+			log_verbose("Unrecognised value for log/%s: %s", cfgname, cv->v.str);
+	}
+
+	return fields;
+}
+
 static void _init_logging(struct cmd_context *cmd)
 {
 	int append = 1;
@@ -330,12 +385,11 @@ static void _init_logging(struct cmd_context *cmd)
 
 	/* Syslog */
 	cmd->default_settings.syslog = find_config_tree_bool(cmd, log_syslog_CFG, NULL);
-	if (cmd->default_settings.syslog != 1)
+	if (cmd->default_settings.syslog)
+		init_syslog(1, DEFAULT_LOG_FACILITY);
+	else
 		fin_syslog();
 
-	if (cmd->default_settings.syslog > 1)
-		init_syslog(cmd->default_settings.syslog);
-
 	/* Debug level for log file output */
 	cmd->default_settings.debug = find_config_tree_int(cmd, log_level_CFG, NULL);
 	init_debug(cmd->default_settings.debug);
@@ -388,6 +442,9 @@ static void _init_logging(struct cmd_context *cmd)
 	init_debug_file_fields(_parse_debug_fields(cmd, log_debug_file_fields_CFG, "debug_file_fields"));
 	init_debug_output_fields(_parse_debug_fields(cmd, log_debug_output_fields_CFG, "debug_output_fields"));
 
+	cmd->default_settings.journal = _parse_log_journal(cmd, log_journal_CFG, "journal");
+	init_log_journal(cmd->default_settings.journal);
+
 	t = time(NULL);
 	ctime_r(&t, &timebuf[0]);
 	timebuf[24] = '\0';
@@ -449,10 +506,10 @@ static int _check_config(struct cmd_context *cmd)
 static const char *_set_time_format(struct cmd_context *cmd)
 {
 	/* Compared to strftime, we do not allow "newline" character - the %n in format. */
-	static const char *allowed_format_chars = "aAbBcCdDeFGghHIjklmMpPrRsStTuUVwWxXyYzZ%";
-	static const char *allowed_alternative_format_chars_e = "cCxXyY";
-	static const char *allowed_alternative_format_chars_o = "deHImMSuUVwWy";
-	static const char *chars_to_check;
+	static const char _allowed_format_chars[] = "aAbBcCdDeFGghHIjklmMpPrRsStTuUVwWxXyYzZ%";
+	static const char _allowed_alternative_format_chars_e[] = "cCxXyY";
+	static const char _allowed_alternative_format_chars_o[] = "deHImMSuUVwWy";
+	const char *chars_to_check;
 	const char *tf = find_config_tree_str(cmd, report_time_format_CFG, NULL);
 	const char *p_fmt;
 	size_t i;
@@ -468,12 +525,12 @@ static const char *_set_time_format(struct cmd_context *cmd)
 				c = *++p_fmt;
 				if (c == 'E') {
 					c = *++p_fmt;
-					chars_to_check = allowed_alternative_format_chars_e;
+					chars_to_check = _allowed_alternative_format_chars_e;
 				} else if (c == 'O') {
 					c = *++p_fmt;
-					chars_to_check = allowed_alternative_format_chars_o;
+					chars_to_check = _allowed_alternative_format_chars_o;
 				} else
-					chars_to_check = allowed_format_chars;
+					chars_to_check = _allowed_format_chars;
 
 				for (i = 0; chars_to_check[i]; i++) {
 					if (c == chars_to_check[i])
@@ -557,11 +614,51 @@ static int _init_system_id(struct cmd_context *cmd)
 	return 1;
 }
 
+static void _init_device_ids_refresh(struct cmd_context *cmd)
+{
+	const struct dm_config_node *cn;
+	const struct dm_config_value *cv;
+	int check_product_uuid = 0;
+	int check_hostname = 0;
+	char path[PATH_MAX];
+	char uuid[128] = { 0 };
+
+	cmd->device_ids_check_product_uuid = 0;
+	cmd->device_ids_check_hostname = 0;
+
+	if (!find_config_tree_bool(cmd, devices_device_ids_refresh_CFG, NULL))
+		return;
+	if (!(cn = find_config_tree_array(cmd, devices_device_ids_refresh_checks_CFG, NULL)))
+		return;
+
+	for (cv = cn->v; cv; cv = cv->next) {
+		if (cv->type != DM_CFG_STRING)
+			continue;
+		if (!strcmp(cv->v.str, "product_uuid"))
+			check_product_uuid = 1;
+		if (!strcmp(cv->v.str, "hostname"))
+			check_hostname = 1;
+	}
+
+	if (check_product_uuid) {
+		const char *sysfs_dir = cmd->device_id_sysfs_dir ?: dm_sysfs_dir();
+		if (dm_snprintf(path, sizeof(path), "%sdevices/virtual/dmi/id/product_uuid", sysfs_dir) < 0)
+			return;
+		if (get_sysfs_value(path, uuid, sizeof(uuid), 0) && uuid[0])
+			cmd->product_uuid = dm_pool_strdup(cmd->libmem, uuid);;
+		if (cmd->product_uuid)
+			cmd->device_ids_check_product_uuid = 1;
+	}
+
+	if (check_hostname && cmd->hostname)
+		cmd->device_ids_check_hostname = 1;
+}
+
 static int _process_config(struct cmd_context *cmd)
 {
 	mode_t old_umask;
 	const char *dev_ext_info_src = NULL;
-	const char *read_ahead;
+	const char *read_ahead, *validate_metadata;
 	struct stat st;
 	const struct dm_config_node *cn;
 	const struct dm_config_value *cv;
@@ -592,13 +689,14 @@ static int _process_config(struct cmd_context *cmd)
 	if (!dm_set_uuid_prefix(UUID_PREFIX))
 		return_0;
 #endif
+	cmd->device_id_sysfs_dir = find_config_tree_str(cmd, devices_device_id_sysfs_dir_CFG, NULL);
 
 	dev_ext_info_src = find_config_tree_str(cmd, devices_external_device_info_source_CFG, NULL);
 
 	if (dev_ext_info_src &&
 	    strcmp(dev_ext_info_src, "none") &&
 	    strcmp(dev_ext_info_src, "udev")) {
-		log_warn("WARNING: unknown external device info source, using none.");
+		log_warn("WARNING: Unknown external device info source, using none.");
 		dev_ext_info_src = NULL;
 	}
 
@@ -606,7 +704,7 @@ static int _process_config(struct cmd_context *cmd)
 		if (udev_init_library_context()) {
 			init_external_device_info_source(DEV_EXT_UDEV);
 		} else {
-			log_warn("WARNING: failed to init udev for external device info, using none.");
+			log_warn("WARNING: Failed to init udev for external device info, using none.");
 			dev_ext_info_src = NULL;
 		}
 	}
@@ -622,7 +720,7 @@ static int _process_config(struct cmd_context *cmd)
 	}
 
 	if (*cmd->proc_dir && !dir_exists(cmd->proc_dir)) {
-		log_warn("WARNING: proc dir %s not found - some checks will be bypassed",
+		log_warn("WARNING: proc dir %s not found - some checks will be bypassed.",
 			 cmd->proc_dir);
 		cmd->proc_dir[0] = '\0';
 	}
@@ -646,6 +744,15 @@ static int _process_config(struct cmd_context *cmd)
 		return 0;
 	}
 
+	cmd->vg_write_validates_vg = 1;
+	if ((validate_metadata = find_config_tree_str(cmd, config_validate_metadata_CFG, NULL))) {
+		if (!strcasecmp(validate_metadata, "none"))
+			cmd->vg_write_validates_vg = 0;
+		else if (strcasecmp(validate_metadata, "full"))
+			log_warn("WARNING: Ignoring unknown validate_metadata setting: %s.",
+				 validate_metadata);
+	}
+
 	/*
 	 * If udev is disabled using DM_DISABLE_UDEV environment
 	 * variable, override existing config and hardcode these:
@@ -686,13 +793,13 @@ static int _process_config(struct cmd_context *cmd)
 	else if (strcmp(cmd->stripe_filler, "error") &&
 		 strcmp(cmd->stripe_filler, "zero")) {
 		if (stat(cmd->stripe_filler, &st)) {
-			log_warn("WARNING: activation/missing_stripe_filler = \"%s\" "
+			log_warn("WARNING: activation/missing_stripe_filler = \"%s\"."
 				 "is invalid,", cmd->stripe_filler);
 			log_warn("         stat failed: %s", strerror(errno));
 			log_warn("Falling back to \"error\" missing_stripe_filler.");
 			cmd->stripe_filler = "error";
 		} else if (!S_ISBLK(st.st_mode)) {
-			log_warn("WARNING: activation/missing_stripe_filler = \"%s\" "
+			log_warn("WARNING: activation/missing_stripe_filler = \"%s\"."
 				 "is not a block device.", cmd->stripe_filler);
 			log_warn("Falling back to \"error\" missing_stripe_filler.");
 			cmd->stripe_filler = "error";
@@ -716,6 +823,7 @@ static int _process_config(struct cmd_context *cmd)
 	init_pv_min_size((uint64_t)pv_min_kb * (1024 >> SECTOR_SHIFT));
 
 	cmd->check_pv_dev_sizes = find_config_tree_bool(cmd, metadata_check_pv_device_sizes_CFG, NULL);
+	cmd->event_activation = find_config_tree_bool(cmd, global_event_activation_CFG, NULL);
 
 	if (!process_profilable_config(cmd))
 		return_0;
@@ -726,6 +834,8 @@ static int _process_config(struct cmd_context *cmd)
 	if (!_init_system_id(cmd))
 		return_0;
 
+	_init_device_ids_refresh(cmd);
+
 	init_io_memory_size(find_config_tree_int(cmd, global_io_memory_size_CFG, NULL));
 
 	return 1;
@@ -1076,7 +1186,7 @@ static int _init_dev_cache(struct cmd_context *cmd)
 	return 1;
 }
 
-#define MAX_FILTERS 11
+#define MAX_FILTERS 10
 
 static struct dev_filter *_init_filter_chain(struct cmd_context *cmd)
 {
@@ -1091,26 +1201,6 @@ static struct dev_filter *_init_filter_chain(struct cmd_context *cmd)
 	 * Update MAX_FILTERS definition above when adding new filters.
 	 */
 
-	/*
-	 * sysfs filter. Only available on 2.6 kernels.  Non-critical.
-	 * Listed first because it's very efficient at eliminating
-	 * unavailable devices.
-	 *
-	 * TODO: I suspect that using the lvm_type and device_id
-	 * filters before this one may be more efficient.
-	 */
-	if (find_config_tree_bool(cmd, devices_sysfs_scan_CFG, NULL)) {
-		if ((filters[nr_filt] = sysfs_filter_create()))
-			nr_filt++;
-	}
-
-	/* internal filter used by command processing. */
-	if (!(filters[nr_filt] = internal_filter_create())) {
-		log_error("Failed to create internal device filter");
-		goto bad;
-	}
-	nr_filt++;
-
 	/* global regex filter. Optional. */
 	if ((cn = find_config_tree_node(cmd, devices_global_filter_CFG, NULL))) {
 		if (!(filters[nr_filt] = regex_filter_create(cn->v, 0, 1))) {
@@ -1143,9 +1233,20 @@ static struct dev_filter *_init_filter_chain(struct cmd_context *cmd)
 	}
 	nr_filt++;
 
+	/*
+	 * sysfs filter. Only available on 2.6 kernels.  Non-critical.
+	 * Eliminates unavailable devices.
+	 * TODO: this may be unnecessary now with device ids
+	 * (currently not used for devs match to device id using sysfs)
+	 */
+	if (find_config_tree_bool(cmd, devices_sysfs_scan_CFG, NULL)) {
+		if ((filters[nr_filt] = sysfs_filter_create(dm_sysfs_dir())))
+			nr_filt++;
+	}
+
 	/* usable device filter. Required. */
-	if (!(filters[nr_filt] = usable_filter_create(cmd, cmd->dev_types, FILTER_MODE_NO_LVMETAD))) {
-		log_error("Failed to create usabled device filter");
+	if (!(filters[nr_filt] = usable_filter_create(cmd, cmd->dev_types))) {
+		log_error("Failed to create usable device filter");
 		goto bad;
 	}
 	nr_filt++;
@@ -1220,7 +1321,7 @@ int init_filters(struct cmd_context *cmd, unsigned load_persistent_cache)
 	init_ignore_lvm_mirrors(find_config_tree_bool(cmd, devices_ignore_lvm_mirrors_CFG, NULL));
 
 	/*
-	 * persisent filter is a cache of the previous result real filter result.
+	 * persistent filter is a cache of the previous result real filter result.
 	 * If a dev is found in persistent filter, the pass/fail result saved by
 	 * the pfilter is used.  If a dev does not existing in the persistent
 	 * filter, the dev is passed on to the real filter, and when the result
@@ -1336,8 +1437,8 @@ static int _init_segtypes(struct cmd_context *cmd)
 	struct segment_type *segtype;
 	struct segtype_library seglib = { .cmd = cmd, .lib = NULL };
 	struct segment_type *(*init_segtype_array[])(struct cmd_context *cmd) = {
-		init_linear_segtype,
 		init_striped_segtype,
+		init_linear_segtype,
 		init_zero_segtype,
 		init_error_segtype,
 		/* disabled until needed init_free_segtype, */
@@ -1462,7 +1563,7 @@ static void _init_rand(struct cmd_context *cmd)
 		return;
 	}
 
-	cmd->rand_seed = (unsigned) time(NULL) + (unsigned) getpid();
+	cmd->rand_seed = (unsigned) ((time(NULL) + getpid()) & 0xffffffff);
 	reset_lvm_errno(1);
 }
 
@@ -1524,63 +1625,6 @@ void destroy_config_context(struct cmd_context *cmd)
 	free(cmd);
 }
 
-/*
- * A "config context" is a very light weight toolcontext that
- * is only used for reading config settings from lvm.conf.
- *
- * FIXME: this needs to go back to parametrized create_toolcontext()
- */
-struct cmd_context *create_config_context(void)
-{
-	struct cmd_context *cmd;
-
-	if (!(cmd = zalloc(sizeof(*cmd))))
-		goto_out;
-
-	strcpy(cmd->system_dir, DEFAULT_SYS_DIR);
-
-	if (!_get_env_vars(cmd))
-		goto_out;
-
-	if (!(cmd->libmem = dm_pool_create("library", 4 * 1024)))
-		goto_out;
-
-	if (!(cmd->mem = dm_pool_create("command", 4 * 1024)))
-		goto out;
-
-	if (!(cmd->pending_delete_mem = dm_pool_create("pending_delete", 1024)))
-		goto_out;
-
-	dm_list_init(&cmd->config_files);
-	dm_list_init(&cmd->tags);
-	dm_list_init(&cmd->hints);
-
-	if (!_init_lvm_conf(cmd))
-		goto_out;
-
-	if (!_init_hostname(cmd))
-		goto_out;
-
-	if (!_init_tags(cmd, cmd->cft))
-		goto_out;
-
-	/* Load lvmlocal.conf */
-	if (*cmd->system_dir && !_load_config_file(cmd, "", 1))
-		goto_out;
-
-	if (!_init_tag_configs(cmd))
-		goto_out;
-
-	if (!(cmd->cft = _merge_config_files(cmd, cmd->cft)))
-		goto_out;
-
-	return cmd;
-out:
-	if (cmd)
-		destroy_config_context(cmd);
-	return NULL;
-}
-
 /* Entry point */
 struct cmd_context *create_toolcontext(unsigned is_clvmd,
 				       const char *system_dir,
@@ -1590,7 +1634,6 @@ struct cmd_context *create_toolcontext(unsigned is_clvmd,
 				       unsigned set_filters)
 {
 	struct cmd_context *cmd;
-	int flags;
 
 #ifdef M_MMAP_MAX
 	mallopt(M_MMAP_MAX, 0);
@@ -1603,8 +1646,6 @@ struct cmd_context *create_toolcontext(unsigned is_clvmd,
 	bindtextdomain(INTL_PACKAGE, LOCALEDIR);
 #endif
 
-	init_syslog(DEFAULT_LOG_FACILITY);
-
 	if (!(cmd = zalloc(sizeof(*cmd)))) {
 		log_error("Failed to allocate command context");
 		return NULL;
@@ -1616,26 +1657,29 @@ struct cmd_context *create_toolcontext(unsigned is_clvmd,
 	cmd->handles_unknown_segments = 0;
 	cmd->hosttags = 0;
 	cmd->check_devs_used = 1;
+	cmd->running_on_valgrind = RUNNING_ON_VALGRIND;
+
 	dm_list_init(&cmd->arg_value_groups);
 	dm_list_init(&cmd->formats);
 	dm_list_init(&cmd->segtypes);
 	dm_list_init(&cmd->tags);
-	dm_list_init(&cmd->hints);
 	dm_list_init(&cmd->config_files);
 	label_init();
 
 	/* FIXME Make this configurable? */
 	reset_lvm_errno(1);
 
-#ifndef VALGRIND_POOL
 	/* Set in/out stream buffering before glibc */
 	if (set_buffering
+	    && !cmd->running_on_valgrind /* Skipping within valgrind execution. */
 #ifdef SYS_gettid
 	    /* For threaded programs no changes of streams */
             /* On linux gettid() is implemented only via syscall */
 	    && (syscall(SYS_gettid) == getpid())
 #endif
 	   ) {
+		int flags;
+
 		/* Allocate 2 buffers */
 		if (!(cmd->linebuffer = malloc(2 * _linebuffer_size))) {
 			log_error("Failed to allocate line buffer.");
@@ -1669,14 +1713,12 @@ struct cmd_context *create_toolcontext(unsigned is_clvmd,
 	} else if (!set_buffering)
 		/* Without buffering, must not use stdin/stdout */
 		init_silent(1);
-#endif
+
 	/*
 	 * Environment variable LVM_SYSTEM_DIR overrides this below.
 	 */
-        if (system_dir)
-		strncpy(cmd->system_dir, system_dir, sizeof(cmd->system_dir) - 1);
-	else
-		strcpy(cmd->system_dir, DEFAULT_SYS_DIR);
+	strncpy(cmd->system_dir, (system_dir) ? system_dir : DEFAULT_SYS_DIR,
+		sizeof(cmd->system_dir) - 1);
 
 	if (!_get_env_vars(cmd))
 		goto_out;
@@ -1866,7 +1908,6 @@ int refresh_toolcontext(struct cmd_context *cmd)
 	_destroy_segtypes(&cmd->segtypes);
 	_destroy_formats(cmd, &cmd->formats);
 
-	devices_file_exit(cmd);
 	if (!dev_cache_exit())
 		stack;
 	_destroy_dev_types(cmd);
@@ -1889,6 +1930,8 @@ int refresh_toolcontext(struct cmd_context *cmd)
 
 	cmd->lib_dir = NULL;
 
+	cmd->lvcreate_vcp = NULL;
+
 	if (!_init_lvm_conf(cmd))
 		return_0;
 
@@ -1984,7 +2027,6 @@ int refresh_toolcontext(struct cmd_context *cmd)
 void destroy_toolcontext(struct cmd_context *cmd)
 {
 	struct dm_config_tree *cft_cmdline;
-	int flags;
 
 	archive_exit(cmd);
 	backup_exit(cmd);
@@ -1995,27 +2037,18 @@ void destroy_toolcontext(struct cmd_context *cmd)
 	_destroy_segtypes(&cmd->segtypes);
 	_destroy_formats(cmd, &cmd->formats);
 	_destroy_filters(cmd);
-	if (cmd->mem)
-		dm_pool_destroy(cmd->mem);
-	devices_file_exit(cmd);
 	dev_cache_exit();
 	_destroy_dev_types(cmd);
 	_destroy_tags(cmd);
 
 	if ((cft_cmdline = remove_config_tree_by_source(cmd, CONFIG_STRING)))
 		config_destroy(cft_cmdline);
-	_destroy_config(cmd);
 
 	if (cmd->cft_def_hash)
 		dm_hash_destroy(cmd->cft_def_hash);
 
-	if (cmd->libmem)
-		dm_pool_destroy(cmd->libmem);
-
-	if (cmd->pending_delete_mem)
-		dm_pool_destroy(cmd->pending_delete_mem);
-#ifndef VALGRIND_POOL
-	if (cmd->linebuffer) {
+	if (!cmd->running_on_valgrind && cmd->linebuffer) {
+		int flags;
 		/* Reset stream buffering to defaults */
 		if (is_valid_fd(STDIN_FILENO) &&
 		    ((flags = fcntl(STDIN_FILENO, F_GETFL)) > 0) &&
@@ -2037,8 +2070,8 @@ void destroy_toolcontext(struct cmd_context *cmd)
 
 		free(cmd->linebuffer);
 	}
-#endif
-	free(cmd);
+
+	destroy_config_context(cmd);
 
 	lvmpolld_disconnect();
 
